import unittest
import numpy as np
import quaternion

from skeletal_animation.core.math.transform import Transform, identity
from skeletal_animation.core.math.constants import HIGH_R_TOL, HIGH_A_TOL
from skeletal_animation.core.model.joint import Joint
from skeletal_animation.core.animate.key import Key
from skeletal_animation.core.animate.curve import Curve
from skeletal_animation.core.animate.jointanimation import JointAnimation

# TODO might want to rewrite the tests about timekeys here 

class TestJointAnimationConstructor(unittest.TestCase):
    def setUp(self):
        self.t = Transform(pos=[0., 1., 0.], orient=quaternion.one)
        self.root_joint = Joint("root", local_bind_transform=self.t)
        self.child_joint = Joint("child", local_bind_transform=identity)
        self.root_joint.add_child(self.child_joint)

        self.translation_curve = Curve([Key(0., [0., 0., 0.]), Key(1., [1., 2., 3.])])
        self.rotation_curve = Curve([Key(0., np.quaternion(1., 0., 0., 0.)), Key(1., np.quaternion(4/5, 3/5, 0., 0.))])
        
    def test_joint_animation_constructor_with_translation_curve(self):
        joint_animation = JointAnimation(joint=self.root_joint, translation_curve=self.translation_curve)
        self.assertEqual(joint_animation.translation_curve, self.translation_curve)
        self.assertEqual(joint_animation.rotation_curve, Curve())

class TestJointAnimationGetLocalTransformWithConstantInterpolation(unittest.TestCase):
    def setUp(self):
        self.t = Transform(pos=[1., 2., 3.], orient=quaternion.one)
        self.root_joint = Joint("root", local_bind_transform=self.t)
        self.child_joint = Joint("child", local_bind_transform=identity)
        self.root_joint.add_child(self.child_joint)
        
        self.child_joint.transform = self.t

        self.translation_curve = Curve([Key(0., [1., 0., 0.]), Key(1., [1., 2., 3.])])
        self.rotation_curve = Curve([Key(0., np.quaternion(1., 0., 0., 0.)), Key(1., np.quaternion(4/5, 3/5, 0., 0.))])

        self.empty_curve = Curve()

    def test_joint_animation_get_local_transform_on_root(self):
        ja = JointAnimation(joint=self.root_joint, translation_curve=self.translation_curve, rotation_curve=self.rotation_curve)
        self.assertEqual(
            ja.get_local_transform(0, "constant"),
            Transform(pos=[1., 0., 0.], orient=np.quaternion(1, 0, 0., 0.))
            )
        self.assertEqual(
            ja.get_local_transform(0.5, "constant"),
            Transform(pos=[1., 0., 0.], orient=np.quaternion(1, 0, 0., 0.))
            )
        self.assertEqual(
            ja.get_local_transform(1, "constant"),
            Transform(pos=[1., 2., 3.], orient=np.quaternion(4/5, 3/5, 0., 0.))
            )
    
    def test_joint_animation_get_local_transform_on_child(self):
        ja = JointAnimation(joint=self.child_joint, translation_curve=self.translation_curve, rotation_curve=self.rotation_curve)
        self.assertEqual(
            ja.get_local_transform(0, "constant"),
            Transform(pos=[1., 0., 0.], orient=np.quaternion(1, 0, 0., 0.))
            )
        self.assertEqual(
            ja.get_local_transform(0.5, "constant"),
            Transform(pos=[1., 0., 0.], orient=np.quaternion(1, 0, 0., 0.))
            )
        self.assertEqual(
            ja.get_local_transform(1, "constant"),
            Transform(pos=[1., 2., 3.], orient=np.quaternion(4/5, 3/5, 0., 0.))
            )

    def test_joint_animation_get_local_transform_on_root_with_no_animation_curve(self):
        ja = JointAnimation(joint=self.root_joint, translation_curve=self.empty_curve, rotation_curve=self.empty_curve)
        self.assertEqual(ja.get_local_transform(0, "constant"), self.t)
        self.assertEqual(ja.get_local_transform(0.5, "constant"), self.t)
        self.assertEqual(ja.get_local_transform(1, "constant"), self.t)
    
    def test_joint_animation_get_local_transform_on_root_with_translation_curve(self):
        ja = JointAnimation(joint=self.root_joint, translation_curve=self.translation_curve, rotation_curve=self.empty_curve)
        self.assertEqual(
            ja.get_local_transform(0, "constant"),
            Transform(pos=[1., 0., 0.], orient=quaternion.one)
            )
        self.assertEqual(
            ja.get_local_transform(0.5, "constant"),
            Transform(pos=[1., 0., 0.], orient=quaternion.one)
            )
        self.assertEqual(
            ja.get_local_transform(1, "constant"),
            Transform(pos=[1., 2., 3.], orient=quaternion.one)
            )
    
    def test_joint_animation_get_local_transform_on_child_with_rotation_curve(self):
        ja = JointAnimation(joint=self.child_joint, translation_curve=self.empty_curve, rotation_curve=self.rotation_curve)
        self.assertEqual(
            ja.get_local_transform(0, "constant"),
            Transform(pos=[0., 0., 0.], orient=np.quaternion(1, 0, 0., 0.))
            )
        self.assertEqual(
            ja.get_local_transform(0.5, "constant"),
            Transform(pos=[0., 0., 0.], orient=np.quaternion(1, 0, 0., 0.))
            )
        self.assertEqual(
            ja.get_local_transform(1, "constant"),
            Transform(pos=[0., 0., 0.], orient=np.quaternion(4/5, 3/5, 0., 0.))
            )

class TestJointAnimationGetLocalTransformWithLinearInterpolation(unittest.TestCase):
    def setUp(self):
        self.t = Transform(pos=[1., 2., 3.], orient=quaternion.one)
        self.root_joint = Joint("root", local_bind_transform=self.t)
        self.child_joint = Joint("child", local_bind_transform=identity)
        self.root_joint.add_child(self.child_joint)
        
        self.child_joint.transform = self.t

        self.translation_curve = Curve([Key(0., [1., 0., 0.]), Key(1., [1., 2., 3.])])

        self.rotation_curve = Curve([Key(0., np.quaternion(1., 0., 0., 0.)), Key(1., np.quaternion(4/5, 3/5, 0., 0.))])

        self.empty_curve = Curve()

    def test_joint_animation_get_local_transform_on_root(self):
        ja = JointAnimation(joint=self.root_joint, translation_curve=self.translation_curve, rotation_curve=self.rotation_curve)
        self.assertEqual(
            ja.get_local_transform(0, "linear"),
            Transform(pos=[1., 0., 0.], orient=np.quaternion(1, 0, 0., 0.))
            )
        self.assertEqual(
            ja.get_local_transform(0.5, "linear"),
            Transform(pos=[1., 1., 1.5], orient=quaternion.slerp(quaternion.one, np.quaternion(4/5, 3/5, 0., 0.), 0, 1, 0.5))
            )
        self.assertEqual(
            ja.get_local_transform(1, "linear"),
            Transform(pos=[1., 2., 3.], orient=np.quaternion(4/5, 3/5, 0., 0.))
            )
    
    def test_joint_animation_get_local_transform_on_root_with_no_animation_curve(self):
        ja = JointAnimation(joint=self.root_joint, translation_curve=self.empty_curve, rotation_curve=self.empty_curve)
        self.assertEqual(ja.get_local_transform(0, "linear"), self.t)
        self.assertEqual(ja.get_local_transform(0.5, "linear"), self.t)
        self.assertEqual(ja.get_local_transform(1, "linear"), self.t)
    
    def test_joint_animation_get_local_transform_on_root_with_translation_curve(self):
        ja = JointAnimation(joint=self.root_joint, translation_curve=self.translation_curve, rotation_curve=self.empty_curve)    
        self.assertEqual(
            ja.get_local_transform(0, "linear"),
            Transform(pos=[1., 0., 0.], orient=quaternion.one)
            )
        self.assertEqual(
            ja.get_local_transform(0.5, "linear"),
            Transform(pos=[1., 1., 1.5], orient=quaternion.one)
            )
        self.assertEqual(
            ja.get_local_transform(1, "linear"),
            Transform(pos=[1., 2., 3.], orient=quaternion.one)
            )
    
    def test_joint_animation_get_local_transform_on_child_with_rotation_curve(self):
        ja = JointAnimation(joint=self.child_joint, translation_curve=self.empty_curve, rotation_curve=self.rotation_curve)    
        self.assertEqual(
            ja.get_local_transform(0, "linear"),
            Transform(pos=[0., 0., 0.], orient=np.quaternion(1, 0, 0., 0.))
            )
        self.assertEqual(
            ja.get_local_transform(0.5, "linear"),
            Transform(pos=[0., 0., 0.], orient=quaternion.slerp(quaternion.one, np.quaternion(4/5, 3/5, 0., 0.), 0, 1, 0.5))
            )
        self.assertEqual(
            ja.get_local_transform(1, "linear"),
            Transform(pos=[0., 0., 0.], orient=np.quaternion(4/5, 3/5, 0., 0.))
            )

class TestJointAnimationGetLocalTransformWithTimewarpCurve(unittest.TestCase):
    def setUp(self):
        # A translation curve that goes from [0,0,0] to [1,1,1] at time 0 and 1 respectively.
        self.translation_curve = Curve(
            [Key(time=0., value=[0,0,0]),
            Key(time=1., value=[1,1,1])]
        )

        # A default time warp curve. The value of the key at time 0 is 0 and the value of the key at time 1 is 1.
        self.timewarp_curve = Curve(
            [Key(time=0., value=0.),
            Key(time=1., value=1.)]
        )
            
        # A JointAnimation object with the above curves.
        self.joint_animation = JointAnimation(
            joint=Joint(),
            translation_curve=self.translation_curve,
            rotation_curve=Curve(),
            timewarp_curve=self.timewarp_curve
        )
        
        self.expected_default_result = [[0,0,0],
        [0.25,0.25,0.25],
        [0.5,0.5,0.5],
        [0.75,0.75,0.75],
        [1,1,1]]

        self.expected_timewarp_result = [[0,0,0],
        [0.1,0.1,0.1],
        [0.2,0.2,0.2],
        [0.6,0.6,0.6],
        [1,1,1]]

    def test_joint_animation_get_local_transform_with_timewarp_curve(self):
        # let us see the result of the get_local_transform method at different times using linear interpolation.
        effective_default_result = []        
        for t in [0., 0.25, 0.5, 0.75, 1.]: 
            effective_default_result.append(self.joint_animation.get_local_transform(t, 'linear').pos)
        self.assertTrue(np.allclose(effective_default_result, self.expected_default_result, rtol=HIGH_R_TOL, atol=HIGH_A_TOL))

        # Now if we add a key to the time warp curve...
        # Adding a time key at time 0.5 with value 0.2.
        self.joint_animation.timewarp_curve.set_key_at(time=0.5, value=0.2)
        # or joint_animation.timewarp_curve.add_key(Key(time=0.5, value=0.2))
        # This means that at time 0.5 of the animation, the translation and rotation curves are evaluated at time 0.2.
        effective_timewarp_result = []
        for t in [0., 0.25, 0.5, 0.75, 1.]: 
            effective_timewarp_result.append(self.joint_animation.get_local_transform(t, 'linear').pos)
        self.assertTrue(np.allclose(effective_timewarp_result, self.expected_timewarp_result, rtol=HIGH_R_TOL, atol=HIGH_A_TOL))



import copy

from skeletal_animation.core.animate import *
from skeletal_animation.utils.other import are_skeletons_similar
from skeletal_animation.core.model import *

class Animator:
    """Controller class to animate a skeleton.

    Attributes:
    ----------
    animation : Animation
        Animation to be played.
    skeleton : Skeleton
        Skeleton to be animated.
    anim_time : float
        Time since the beginning of the animation.
    kps : int
        Keys per second of the animation. Corresponds to the speed at which the animation is played.
    duration : float
        Duration of the animation in seconds.
    interpolation_algo : str
        Algorithm to be used to interpolate between the keys of the animation.
    is_playing : bool
        Boolean to determine whether the animation is playing or paused.
        Only used when the animator is played by the gui script (frame.py).
    """

    def __init__(self, animation, skeleton, kps=30, interpolation_algo = "linear"):
        """**Default constructor for the Animator class.**

        Parameters:
        ----------
        animation : Animation
            Animation to be played.
        skeleton : Skeleton
            Skeleton to be animated.    
        kps : int, optional
            Keys per second of the animation, by default 30.
        interpolation_algo : str, optional
            Interpolation algorithm to be used, by default "linear".
            Possible values: "constant" and "linear".
        """
        self.animation = animation 
        self.skeleton = skeleton
        self.anim_time = 0
        self.kps = kps
        self.duration = self.compute_animation_duration()
        self.interpolation_algo = interpolation_algo
        self.is_playing = True

    def compute_animation_duration(self):
        """Computes the duration of the animation in seconds.

        Called by the constructor.

        Returns:
        ----------        
        float
            The duration of the animation in seconds.
        """
        return self.animation.max_key_time / self.kps


    def generate_pose(self):
        """Generates a pose dictionary.

        The pose dictionary is of the form:
        {int: Transform}
        The integer key is the id of a joint in the skeleton and
        the transform value is the local animation transform of the joint.

        The local animation transform is the transform that must be applied to the joint's local bind transform
        to obtain the model transform of the joint at a specific time in the animation.

        Returns:
        ----------
        dict (int: Transform)
            Pose dictionary generated.
            Where the integer key is the id of a joint in the skeleton and 
            the transform value is the local animation transform of the joint.

        See Also:
        ----------
        joint.forward_kinematics: method that computes the model transform of the joints
        using the pose dictionary.
        jointanimation.get_local_transform: method used to compute the local animation
        transform contained in the pose dictionary.
        """
        time = (self.anim_time / self.duration) % 1.0
        pose_dict = {}

        for joint_animation in self.animation.joint_animations:
            
            joint = joint_animation.joint                        
            
            local_transform = joint_animation.get_local_transform(time, self.interpolation_algo)

            pose_dict[joint.id] = local_transform      

        return pose_dict  
        
    def update(self, time, verbose=False):
        """Updates the animation time.

        Parameters:
        ----------
        time : float
            Time to be added to the animation time.
            Is affected by the speed of the animation in kps.
        verbose : bool, optional
            If set to true the effective time added to the animation time 
            will be printed. by default False.
        """
        if verbose:
            print("_dt:", time * (self.kps / 30))
        self.anim_time += time * (self.kps / 30)
    

    def play(self, play_duration = 1., fps=30., verbose=False):
        """Plays the animation for a given duration.

        This method is used to simulate the animation without rendering it.
        The animation is played for a simulated duration in seconds.
        The amount of step to compute per second is given by the fps parameter.

        Parameters:
        ----------
        play_duration : float, optional
            The length of time the animation is left running in seconds, by default 1.
        fps : float, optional
            The amount of steps to compute per second, by default 30.
        verbose : bool, optional
            If set to true the model transform of each joint will be printed at every step,
            by default False.    
        """
        while self.anim_time <= play_duration:
            self.step(1/fps, verbose)


    def step(self, time, verbose=False):
        """Steps the animation for a given time.

        Calls the update method to update the animation time.
        Calls the generate_pose method to generate the pose dictionary.
        Then calls the compute_pose method of the skeleton to update the model transform
        of every joint using forward kinematics.

        Parameters:
        ----------
        time : float
            Time to step the animation for.
            Negative values step the animation backwards.
        verbose : bool, optional
            If set to true the current time of the animation and the model transform
            of each joint will be printed, by default False.        
        """
        self.update(time, verbose)
        pose_dict = self.generate_pose()
        self.skeleton.compute_pose(pose_dict)
        if verbose:
            print("_Time:", self.anim_time)
            for joint in self.skeleton.as_joint_list():
                print("Joint:", joint.name, "Transform:", joint.model_transform)

    def translate(self, x, y, z):
        """Translates the animation by a given vector.

        The translation is only applied to the root joint of the skeleton.
        Every key in the translation curve for the root is translated by the given vector.

        Parameters:
        ----------
        x : float
            Amount of the translation in the x direction.
        y : floatthe 
            Amount of the translation in the y direction.
        z : floatthe 
            Amount of the translation in the z direction.
        """
        animation = self.animation
        root_animation = animation.joint_animations[0]
        for key in root_animation.translation_curve.keys:
            key.value = [key.value[0] + x, key.value[1] + y, key.value[2] + z]

    def clip(self, other):
        """Clips the other animation at the end of the current animation.

        Join the two animations together appending the keys of the other animation
        to the end of the current animation.
        This creates a new animation that is the concatenation of the two.
        The duration of the new animation is the sum of the duration of the two animations.
        Curves are normalized to keep the time values between 0 and 1.
        
        Parameters:
        ----------
        other : Animation
            Animation to be clipped to the current animation.
        
        Returns:
        ----------
        Animator
            New animator containing the concatenated animation.

        Notes
        -----
        The initial animator objects self and other are not modified.
        """
        if not are_skeletons_similar(self.skeleton, other.skeleton):
            raise ValueError("Cannot generate blend between the two given animators. Skeletons are not identical.")

        self_animation = copy.deepcopy(self.animation)
        other_animation = copy.deepcopy(other.animation)

        total_duration = self_animation.max_key_time + other_animation.max_key_time
        self_duration_ratio = self_animation.max_key_time / total_duration

        new_joint_animations = []
        
        for ja_id in range(len(self.animation.joint_animations)):
            self_joint_animation = self_animation.joint_animations[ja_id]
            other_joint_animation = other_animation.joint_animations[ja_id]
            new_joint_animation = JointAnimation(self_joint_animation.joint,
                                                rotation_curve=Curve(),
                                                translation_curve=Curve(),
                                                timewarp_curve=Curve())

            # translation
            self_translation_curve = self_joint_animation.translation_curve
            other_translation_curve = other_joint_animation.translation_curve
            if self_translation_curve is not None:
                self_translation_curve.normalize_times(total_duration / self_animation.max_key_time)
            if other_translation_curve is not None:
                other_translation_curve.normalize_times(total_duration / other_animation.max_key_time)
            for key in self_translation_curve.keys:
                new_joint_animation.translation_curve.set_key_at(key.time, key.value)
            for key in other_translation_curve.keys:
                new_joint_animation.translation_curve.set_key_at(key.time + self_duration_ratio, key.value)

            # rotation
            self_rotation_curve = self_joint_animation.rotation_curve
            other_rotation_curve = other_joint_animation.rotation_curve
            if self_rotation_curve is not None:
                self_rotation_curve.normalize_times(total_duration / self_animation.max_key_time)
            if other_rotation_curve is not None:
                other_rotation_curve.normalize_times(total_duration / other_animation.max_key_time)
            for key in self_rotation_curve.keys:
                new_joint_animation.rotation_curve.set_key_at(key.time, key.value)
            for key in other_rotation_curve.keys:
                new_joint_animation.rotation_curve.set_key_at(key.time + self_duration_ratio, key.value)

            # timewarp
            self_timewarp_curve = self_joint_animation.timewarp_curve
            other_timewarp_curve = other_joint_animation.timewarp_curve
            if self_timewarp_curve is not None:
                self_timewarp_curve.normalize_times(total_duration / self_animation.max_key_time)
                self_timewarp_curve.normalize_values(total_duration / self_animation.max_key_time)
            if other_timewarp_curve is not None:
                other_timewarp_curve.normalize_times(total_duration / other_animation.max_key_time)
                other_timewarp_curve.normalize_values(total_duration / other_animation.max_key_time)
            for key in self_timewarp_curve.keys:
                new_joint_animation.timewarp_curve.set_key_at(key.time, key.value)
            for key in other_timewarp_curve.keys:
                new_joint_animation.timewarp_curve.set_key_at(key.time + self_duration_ratio, key.value + self_duration_ratio)        

            new_joint_animations.append(new_joint_animation)    

        new_animation = Animation(new_joint_animations, total_duration)
        new_skeleton = copy.deepcopy(self.skeleton)

        new_animator = Animator(new_animation, new_skeleton, self.kps, self.interpolation_algo)

        return new_animator       

    def crop(self, start_time, end_time):
        """Crops the animation to be between the given start and end times.

        The resulting animation contains all the keys that are between the given start and end times.
        The duration of the resulting animation is the difference between the given start and end times.
        Curves are normalized to keep the time values between 0 and 1.

        Parameters:
        ----------
        start_time : float
            Start time of the cropped animation.
        end_time : float
            End time of the cropped animation.
        
        Returns:
        ----------
        Animator
            New animator containing the cropped animation.

        Raises:
        ----------
        ValueError
            If the start time is greater or equal to the end time.
        
        Notes
        -----
        The initial animator object is not modified.
        """
        if start_time >= end_time:
            raise ValueError("Start time must be strictly less than end time.")

        self_animation = copy.deepcopy(self.animation)

        new_joint_animations = []

        for ja in self_animation.joint_animations:
            new_joint_animation = JointAnimation(ja.joint,
                                                rotation_curve=Curve(),
                                                translation_curve=Curve(),
                                                timewarp_curve=Curve()) 
            new_joint_animation.timewarp_curve.remove_key_at(0.0)           
            new_joint_animation.timewarp_curve.remove_key_at(1.0)

            # translation
            self_translation_curve = ja.translation_curve
            if self_translation_curve is not None:
                for key in self_translation_curve.keys:
                    start_translation_value = self_translation_curve.get_lerp_value(start_time)
                    end_translation_value = self_translation_curve.get_lerp_value(end_time)
                    new_joint_animation.translation_curve.set_key_at(0., start_translation_value)
                    new_joint_animation.translation_curve.set_key_at(end_time - start_time, end_translation_value)
                
                    if key.time <= end_time and key.time >= start_time:
                        new_joint_animation.translation_curve.set_key_at(key.time - start_time, key.value)

            # rotation
            self_rotation_curve = ja.rotation_curve
            if self_rotation_curve is not None:
                for key in self_rotation_curve.keys:
                    start_rotation_value = self_rotation_curve.get_lerp_value(start_time)
                    end_rotation_value = self_rotation_curve.get_lerp_value(end_time)
                    new_joint_animation.rotation_curve.set_key_at(0., start_rotation_value)
                    new_joint_animation.rotation_curve.set_key_at(end_time - start_time, end_rotation_value)
                
                    if key.time <= end_time and key.time >= start_time:
                        new_joint_animation.rotation_curve.set_key_at(key.time - start_time, key.value)

            # timewarp
            self_timewarp_curve = ja.timewarp_curve
            if self_timewarp_curve is not None:
                for key in self_timewarp_curve.keys:
                    start_timewarp_value = self_timewarp_curve.get_lerp_value(start_time)
                    end_timewarp_value = self_timewarp_curve.get_lerp_value(end_time)
                    new_joint_animation.timewarp_curve.set_key_at(0., start_timewarp_value)
                    new_joint_animation.timewarp_curve.set_key_at(end_time - start_time, end_timewarp_value)
                
                    if key.time <= end_time and key.time >= start_time:
                        new_joint_animation.timewarp_curve.set_key_at(key.time - start_time, key.value)

            new_joint_animations.append(new_joint_animation)

        for new_ja in new_joint_animations:
            new_timewarp_curve = new_ja.timewarp_curve
            min_timewarp_value = 1.0
            max_timewarp_value = 0.0     

            for key in new_timewarp_curve.keys:
                if key.value < min_timewarp_value:
                    min_timewarp_value = key.value
                if key.value > max_timewarp_value:
                    max_timewarp_value = key.value

            for key in new_timewarp_curve.keys:
                key.value -= min_timewarp_value          

            new_ja.translation_curve.normalize_times(end_time - start_time)
            new_ja.rotation_curve.normalize_times(end_time - start_time)
            new_ja.timewarp_curve.normalize_times(end_time - start_time)
            new_ja.timewarp_curve.normalize_values(max_timewarp_value - min_timewarp_value)

        cropped_duration = (end_time - start_time) * self_animation.max_key_time

        new_animation = Animation(new_joint_animations, cropped_duration)
        new_skeleton = copy.deepcopy(self.skeleton)

        return Animator(new_animation, new_skeleton, self.kps, self.interpolation_algo)

    # Utility functions for live coding.
    def pause(self):
        """Pauses the animation.

        Sets the is_playing flag to False.
        """
        self.is_playing = False
    
    def resume(self):
        """Resumes the animation.

        Sets the is_playing flag to True.
        """
        self.is_playing = True

    def add_timekey(self, real_time, key_time):
        """Adds a timekey to the animation.

        The same timekey is added to the timewarp curve of every joint animation
        in the animator's animation.
        More information on timewarp curve and timekeys can be found in the documentation of the
        JointAnimation class.

        Parameters:
        ----------
        real_time : float
            The time of the timekey.
        key_time : float
            The value of the timekey.
        """
        for joint_animation in self.animation.joint_animations:
            joint_animation.timewarp_curve.set_key_at(real_time, key_time)

    def remove_timekey_at_time(self, time):
        """Removes a timekey from the animation at a given time.

        Remove the timekey from the timewarp curve of every joint animation in the animator's animation
        at the given time.
        More information on timewarp curve and timekeys can be found in the documentation of the
        JointAnimation class.

        Parameters:
        ----------
        time : float
            The time at which the timekeys are removed.
        """
        for joint_animation in self.animation.joint_animations:
            joint_animation.timewarp_curve.remove_key_at(time)



import unittest
import numpy as np
import quaternion

from skeletal_animation.core.math.transform import Transform, identity
from skeletal_animation.core.math.constants import HIGH_R_TOL, HIGH_A_TOL

class TestTransformDefaultConstructor(unittest.TestCase):
    def setUp(self):
        self.t1 = identity
        self.t2 = Transform([1,2,3], np.quaternion(2/3, 0, 2/3, 1/3)) # arbitrary values
        self.m2 = np.array([[-1/9, -4/9, 8/9, 1], # expected values for self.t2 as a matrix
                            [4/9, 7/9, 4/9, 2],
                            [-8/9, 4/9, 1/9, 3],
                            [0, 0, 0, 1]])

    def test_default_constructor_with_default_params(self):        
        self.assertTrue(np.array_equal(self.t1.pos, np.array([0,0,0])))
        self.assertEqual(self.t1.orient, quaternion.one)
        self.assertTrue(np.allclose(self.t1.as_matrix(), np.eye(4), rtol=HIGH_R_TOL, atol=HIGH_A_TOL))
    
    def test_default_constructor_with_unit_quaternion(self):
        self.assertTrue(np.array_equal(self.t2.pos, np.array([1,2,3])))
        self.assertEqual(self.t2.orient, np.quaternion(2/3, 0, 2/3, 1/3))
        self.assertTrue(np.allclose(self.t2.as_matrix(), self.m2, rtol=HIGH_R_TOL, atol=HIGH_A_TOL))
    
    def test_default_constructor_with_non_unit_quaternion(self):
        q = np.quaternion(1,2,3,4)

        with self.assertWarns(UserWarning):
            t = Transform([1,2,3], q)
        self.assertTrue(np.allclose(t.as_matrix()[0:3,0:3], quaternion.as_rotation_matrix(q), rtol=HIGH_R_TOL, atol=HIGH_A_TOL))

class TestTransformFromMatrixConstructor(unittest.TestCase):
    def setUp(self):
        self.t1 = identity
        self.t2 = Transform([1,2,3], np.quaternion(2/3, 0, 2/3, 1/3)) # arbitrary values
        self.m2 = np.array([[-1/9, -4/9, 8/9, 1], # expected values for self.t2 as a matrix
                            [4/9, 7/9, 4/9, 2],
                            [-8/9, 4/9, 1/9, 3],
                            [0, 0, 0, 1]])

    def test_from_identity_matrix(self):
        t = Transform.from_matrix(np.eye(4))
        self.assertEqual(self.t1, t)

    def test_from_matrix(self):
        t = Transform.from_matrix(self.m2)
        self.assertEqual(self.t2, t)
    
class TestTransformAsMatrix(unittest.TestCase):
    def setUp(self):
        self.t1 = identity
        self.t2 = Transform([1,2,3], np.quaternion(2/3, 0, 2/3, 1/3)) # arbitrary values
        self.m2 = np.array([[-1/9, -4/9, 8/9, 1], # expected values for self.t2 as a matrix
                            [4/9, 7/9, 4/9, 2],
                            [-8/9, 4/9, 1/9, 3],
                            [0, 0, 0, 1]])

    def test_as_identity_matrix(self):
        np.allclose(self.t1.as_matrix(), np.eye(4), rtol=HIGH_R_TOL, atol=HIGH_A_TOL)

    def test_as_matrix(self):
        np.allclose(self.t2.as_matrix(), self.m2, rtol=HIGH_R_TOL, atol=HIGH_A_TOL)

class TestTransformNormalizeQuaternion(unittest.TestCase):
    def test_normalize_quaternion(self):
        with self.assertWarnsRegex(UserWarning, "Transform built with non unit quaternion."):
            self.t2 = Transform(orient=np.quaternion(1, 1, 0, 0)) # arbitrary values, non unit quaternion
        self.t2.normalize_quaternion()
        self.assertTrue(np.allclose(quaternion.as_float_array(self.t2.orient), [np.sqrt(2)/2, np.sqrt(2)/2, 0, 0]))


class TestTransformMultiplyBy(unittest.TestCase):
    def setUp(self):
        self.t1 = identity
        self.t2 = Transform([1,2,3], np.quaternion(2/3, 0, 2/3, 1/3)) # arbitrary values
        self.m2 = np.array([[-1/9, -4/9, 8/9, 1], # expected values for self.t2 as a matrix
                            [4/9, 7/9, 4/9, 2],
                            [-8/9, 4/9, 1/9, 3],
                            [0, 0, 0, 1]])

        self.p = [0,2,1]  # arbitrary values,
        self.q = np.quaternion(2/3, 0, 2/3, 1/3) # arbitrary values,

        self.t4 = Transform(pos=self.p)
        self.t5 = Transform(orient=self.q)
        self.t6 = Transform(self.p,self.q)

    def test_multiply_by_identity(self):
        self.assertTrue(np.allclose(self.t2.multiply_by(self.t1).as_matrix(), self.t2.as_matrix(), rtol=HIGH_R_TOL, atol=HIGH_A_TOL))
    
    def test_multiply_by_non_unit(self):

        with self.assertWarnsRegex(UserWarning, "Transform built with non unit quaternion."):
            t3 = Transform([-3, 0.5, -0.02], np.quaternion(-1/2, -5/16, -1/16, 1/8)) # arbitrary values, non unit quaternion

        """
        t3.as_matrix():
        [[ 0.89361702  0.44680851 -0.04255319  -3        ]
        [-0.23404255  0.38297872 -0.89361702  0.5        ]
        [-0.38297872  0.80851064  0.44680851  -0.02       ]
        [0            0           0           1        ]]
        """

        tm = np.array([[-0.33569739776548, 0.49881796999947, 0.79905437221380, 1.0933333333333322],
                        [0.0449172611270, 0.855791959964, -0.51536642882118, 1.0466666666666812],
                        [-0.94089834220898, -0.1371158400067, -0.30969267222359, 5.8866666666666678],
                        [0, 0, 0, 1]]) # Result given by online calculator : https://matrix.reshish.com/fr/multCalculation.php
                                                                    # and confirmed by WolframAlpha : https://www.wolframalpha.com/

        with self.assertWarnsRegex(UserWarning, "Transform multiplication attempted with non unit quaternions."):
            self.assertTrue(np.allclose((self.t2.multiply_by(t3)).as_matrix(), tm, rtol=HIGH_R_TOL, atol=HIGH_A_TOL))
    
    def test_multiply_combine_translation_and_rotation(self):
        self.assertEqual(self.t5.multiply_by(self.t4), self.t6)
    
    def test_multiply_by_translation(self):
        self.assertEqual(self.t6, self.t5.translate(self.p))

    def test_multiply_by_rotation(self):        
        self.assertEqual(self.t6, self.t4.rotate_quaternion(self.q))
        
"""
Data formated for easy import into online calculator https://matrix.reshish.com/fr/multCalculation.php:

-0.11111111111111 -0.44444444444444 0.88888888888889 1
0.44444444444444 0.77777777777778 0.44444444444444 2
-0.88888888888889 0.44444444444444 0.11111111111111 3
0 0 0 1

-0.111111	0.4444444	-0.8888883	-1.666666
-0.444444	0.7777777	0.4444444	-3.333333
0.8888888	0.4444444	0.1111111	-0.3333333
0 0 0 1
"""

class TestTransformLerp(unittest.TestCase):
    def setUp(self):
        self.t1 = identity
        self.p = [1,2,3] # arbitrary values
        self.q = np.quaternion(2/3, 0, 2/3, 1/3) # arbitrary values
    
    def test_lerp_pos(self):
        t2 = Transform(pos=self.p)
        t = Transform.lerp(self.t1, t2, 0.25) # arbitrary time
        self.assertTrue(np.allclose(t.pos, [0.25, 0.5, 0.75], rtol=HIGH_R_TOL, atol=HIGH_A_TOL))
    
    def test_lerp_orient(self):
        t2 = Transform(orient=self.q)
        t = Transform.lerp(self.t1, t2, 0.25) # arbitrary time
        self.assertTrue(np.allclose(quaternion.as_float_array(t.orient), [0.9779751860797075, 0, 0.1866859082252575, 0.09334295411262875], rtol=HIGH_R_TOL, atol=HIGH_A_TOL))



import warnings
import copy 

from skeletal_animation.core.math import *
from skeletal_animation.core.model import *
from skeletal_animation.core.animate import *
from skeletal_animation.utils.other import are_skeletons_similar


def generate_blend_animator(first_animator, second_animator, blend_curve=Curve()):
    """Generates a new animator that is the blend of the two given animators according to a blend curve.

    The blend curve allows the weight to vary over time to allow for transitions between animations.
    The blend curve is a Curve object with keys that give the weight of the blend 
    at a given time. Weight values between keys are linearly interpolated.

    A weight of 0.0 means the resulting animator is the first animator.
    A weight of 1.0 means the resulting animator is the second animator.
    Any weight in between means the resulting animator is a blend of the two animators.
    The lower the weight, the closer the resulting animation is to the first animator's animation. 

    If the blend curve has no keys, the resulting animator is a constant  blend of the two animators
    with a weight of 0.5.

    Parameters:
    -----------
    first_animator : Animator
        The first animator to blend.
    second_animator : Animator
        The second animator to blend.
    blend_curve : Curve, optional
        Curve that describes the weight value of the blend over the course of the animation. 
        By default, a constant curve with weight 0.5.
    
    Returns:
    --------
    Animator
        The animator that is the blend of the two given animators. None if the two animators are not compatible.

    Warns:
    ------
    UserWarning
        If the two animators are not compatible.
    UserWarning
        If the blend curve is empty.
    """
    if not are_skeletons_similar(first_animator.skeleton, second_animator.skeleton):
        warnings.warn("Cannot generate blend between the two given animators. Skeletons are not identical. Returning None.")
        return None
    
    first_animation = copy.deepcopy(first_animator.animation)
    second_animation = copy.deepcopy(second_animator.animation)

    if blend_curve.is_empty():
        warnings.warn("Blend curve is empty. Using default curve with a constant weight of 0.5.", stacklevel=2)
        blend_curve.set_key_at(0, 0.5)

    new_joint_animations = []

    for ja_id in range(len(first_animation.joint_animations)):
        first_joint_animation = first_animation.joint_animations[ja_id]
        second_joint_animation = second_animation.joint_animations[ja_id]
        new_joint_animation = JointAnimation(joint=first_joint_animation.joint
            , translation_curve=Curve()
            , rotation_curve=Curve()
            , timewarp_curve=Curve())
            
        # Translation curve
        frame_times = [key.time for key in first_joint_animation.translation_curve.keys]
        for frame_time in frame_times:
            blend_factor = blend_curve.get_lerp_value(frame_time)

            first_translation_value = first_joint_animation.translation_curve.get_key_at(frame_time).value
            second_translation_value = second_joint_animation.translation_curve.get_lerp_value(frame_time)
            if first_translation_value is None:
                if second_translation_value is None:
                    new_translation_key_value = None
                else:
                    new_translation_key_value = second_translation_value
            else:
                if second_translation_value is None:
                    new_translation_key_value = first_translation_value
                else:
                    new_translation_key_value = Curve.lerp(first_translation_value,
                                                        second_translation_value,
                                                        blend_factor)

            if new_translation_key_value is not None:
                new_joint_animation.translation_curve.set_key_at(frame_time, new_translation_key_value)

        # Rotation curve
        frame_times = [key.time for key in first_joint_animation.rotation_curve.keys]
        for frame_time in frame_times:
            blend_factor = blend_curve.get_lerp_value(frame_time)

            first_rotation_value = first_joint_animation.rotation_curve.get_key_at(frame_time).value
            second_rotation_value = second_joint_animation.rotation_curve.get_lerp_value(frame_time)
            if first_rotation_value is None:
                if second_rotation_value is None:
                    new_rotation_key_value = None
                else:
                    new_rotation_key_value = second_rotation_value
            else:
                if second_rotation_value is None:
                    new_rotation_key_value = first_rotation_value
                else:
                    new_rotation_key_value = Curve.lerp(first_rotation_value,
                                                        second_rotation_value,
                                                        blend_factor)

            if new_rotation_key_value is not None:
                new_joint_animation.rotation_curve.set_key_at(frame_time, new_rotation_key_value)
            
        # Timewarp curve
        frame_times = [key.time for key in first_joint_animation.timewarp_curve.keys]
        for frame_time in frame_times:
            blend_factor = blend_curve.get_lerp_value(frame_time)

            first_timewarp_value = first_joint_animation.timewarp_curve.get_key_at(frame_time).value
            second_timewarp_value = second_joint_animation.timewarp_curve.get_lerp_value(frame_time)
            if first_timewarp_value is None:
                if second_timewarp_value is None:
                    new_timewarp_key_value = None
                else:
                    new_timewarp_key_value = second_timewarp_value
            else:
                if second_timewarp_value is None:
                    new_timewarp_key_value = first_timewarp_value
                else:
                    new_timewarp_key_value = Curve.lerp(first_timewarp_value,
                                                        second_timewarp_value,
                                                        blend_factor)

            if new_timewarp_key_value is not None:
                new_joint_animation.timewarp_curve.set_key_at(frame_time, new_timewarp_key_value)

        new_joint_animations.append(new_joint_animation)

    new_animation = Animation(joint_animations=new_joint_animations, max_key_time=first_animation.max_key_time)
    new_skeleton = copy.deepcopy(first_animator.skeleton)

    new_animator = Animator(
        new_animation,
        new_skeleton,
        first_animator.kps,
        first_animator.interpolation_algo)

    return new_animator
    





class Skeleton:
    """A class to represent a skeleton (or rig).

    Attributes
    ----------
    name : str
        Name given to the skeleton. Not necessarily unique.
    root : Joint
        Root joint of the hierarchy for the skeleton.    
    """

    def __init__(self, root, name="new skeleton"):
        """**Default constructor for the Skeleton class.**
        
        Parameters
        ----------
        root : Joint
            Root joint of the hierarchy for the skeleton.
        name : str, optional
            Name of the skeleton, by default "new skeleton".        
        """
        self.name = name
        self.root = root
    
    
    def as_joint_list(self):
        """Returns the skeleton as a list of joints.

        Note
        ----
        The list is in depth-first order starting from the root.
        
        Returns
        -------
        (Joint) list
            List of joints in the skeleton.
        """
        adjacency_list = []

        # Depth-first search
        def dfs(joint):
            adjacency_list.append(joint)
            for child in joint.children:
                dfs(child)

        dfs(self.root)
        return adjacency_list
    
    
    def get_joint_by_id(self, id):
        """Returns the joint with the given id.
        
        Parameters
        ----------
        id : int
            Id of the joint to return.
        
        Returns
        -------
        Joint
            Joint with the given id. None if not found.
        """
        for joint in self.as_joint_list():
            if joint.id == id:
                return joint
        return None

    
    def compute_bind_pose(self):
        """Computes the initial model transform (i.e. model bind transform) for each joint in the skeleton's joint hierarchy.
        """
        self.root.forward_kinematics(None)

    def compute_pose(self, pose_dict):
        """Computes the model transform for each joint in the skeleton's joint hierarchy.
        
        Parameters
        ----------
        pose_dict : Dict[int, Transform]
            Dictionary with joint id as keys and their respective transform as values.
            Dictionary containing the local transforms to be applied to each joint in the skeleton for it 
            to be in the current pose. It is obtained from the animation data stored in curves. 
            See generate_pose() method in the Animator class.
        """
        self.root.forward_kinematics(pose_dict)
    
    def __repr__(self):
        s = "Name: " + self.name + "\nJoints:\n(root)\n"
        s += ",\n".join(joint.__repr__() for joint in self.as_joint_list())        
        return s



import warnings
import numpy as np
import quaternion

from skeletal_animation.core.math.constants import VERY_HIGH_A_TOL, VERY_HIGH_R_TOL
from skeletal_animation.core.math.transform import Transform
from skeletal_animation.core.animate.key import Key

class Curve:
    """Class to represent an animation curve.

    It is a list of keys that are used to animate a value over time.
    It contains a list of keys as its single attribute and provides
    methods to manipulate the keys and interpolate between them.

    The keys are sorted by increasing time when added or edited via the 
    add_key() and set_key_at() methods.

    All keys in the animation curve must be of the same type.
    A curve object can technically hold keys of any type.
    However, the type of keys supported for linear interpolation are:
    - int
    - float
    - list or array of floats
    - numpy.quaternion
    - Transform

    Curves do not support keys with a negative time.

    Attributes:
    ----------
    keys : (Key) list
        The keys that make up the animation curve.
        They are sorted by ascending time whenever a new one is added.

    """
    def __init__(self, keys=[]):
        """**Default constructor for the Curve class.**

        Parameters:
        ----------
        keys : (Key) list, optional
            The keys of the animation curve, by default empty list.
        """
        self.keys = []
        for key in keys:
            self.add_key(key)
    
    def is_empty(self):
        """Returns whether the animation curve is empty.

        Returns:
        ----------
        bool
            True if the animation curve is empty, False otherwise.
        """
        return len(self.keys) == 0
    
    def get_key_count(self):
        """Returns the number of keys in the animation curve.

        Returns:
        ----------
        int
            The number of keys in the animation curve.
        """
        return len(self.keys)

    def get_max_key_time(self):
        """Returns the maximum key time of the animation curve.

        Returns:
        ----------
        float
            The maximum time of the animation curve.
        """
        if self.keys:
            return max([key.time for key in self.keys])
        else:
            return 0.0

    def normalize_times(self, max_time):
        """Normalizes the animation curve. 
        Effectively divides the time of each key by the given max time.
        This used to restrict the time of the keys to a specific range.
        Generally, 0.0 to 1.0.

        Parameters:
        ----------
        max_time : float
            The time to normalize the animation curve by.
        """
        if max_time == 0:
            # Max key time is 0, no normalization needed.
            pass            
        else:
            for key in self.keys:
                key.time /= max_time
    
    def normalize_values(self, ratio):
        """Normalizes the values of the animation curve.
        Effectively multiplies the value of each key by the given ratio.
        This used to restrict the values of the keys to a specific range.

        Note:
        ----------
        This method is only supported for keys of type int, float.

        Parameters:
        ----------
        ratio : float
            The ratio to normalize the animation curve by.

        Warns:
        ----------
        Warning
            If the animation curve contains keys of type other than int or float.

        Raises:
        ----------
        ZeroDivisionError
            If the ratio is 0.
        """
        if ratio == 0:
            raise ZeroDivisionError("Invalid ratio. Cannot normalize by 0.")

        if not self.is_empty():
            if self.keys[0].get_type() in [int, float, np.float32, np.float64]:
                for key in self.keys:
                    key.value /= ratio
            else:
                warnings.warn(f"Cannot normalize values of type {str(self.keys[0].get_type())}. Only int and float supported.", stacklevel=2)

    def add_key(self, key):
        """Adds a key to the animation curve.

        Note:
        ----------
        this method uses the get_key_at() method to check if a key already exists at the given time.
        See the documentation of that method for more details about the tolerance values of the check.

        Raises:
        ----------
        ValueError
            If the time of the key is negative.
        ValueError
            If type of the key does not match the type of the keys in the curve.         
        ValueError
            If a key with the same time already exists in animation curve.
            To edit an existing key, use the set_key_at() method instead.

        Warns:
        ----------
        UserWarning
            If the time of the key is greater than 1.0. 
        
        Parameters:
        ----------
        key : Key
            The key to be added to the animation curve.
        """
        if key.time < 0.0:
            raise ValueError(f"Invalid key time {key.time}. Key time must be greater than 0.")
        if key.time > 1.0:
            warnings.warn(f"Given key time {key.time}. Key time should be less than 1 for animation \
            purposes. Curve normalization is required.", stacklevel=2)

        if self.keys:
            if key.get_type() != self.keys[0].get_type():
                if key.get_type() in [int, float, np.float32, np.float64] and self.keys[0].get_type() in [int, float, np.float32, np.float64]:
                    pass
                elif key.get_type() in [np.ndarray, list] and self.keys[0].get_type() in [np.ndarray, list]:
                    pass                    
                else:
                    raise ValueError(f"Key type ({str(key.get_type())}) does not match \
                    the type of the first key of the animation curve ({str(self.keys[0].get_type())}).")

            if self.get_key_at(key.time) is not None:
                raise ValueError(f"Key at the given time {str(key.time)} already exists in animation curve. \
                Use set_key_at() to change the value of an existing key.")
            else:
                self.keys.append(key)
                self.keys.sort(key=lambda key: key.time)
        else:
            self.keys.append(key)     

    def set_key_at(self, time, value):
        """Sets the value of the key at the given time.
        If no key exists at the given time, a new key is added
        via the add_key() method.

        Note:
        ----------
        this method uses the get_key_at() method to check if a key already exists at the given time.
        See the documentation of that method for more details about the tolerance values of the check.

        Parameters:
        ----------
        time : float
            The time of the key to be set.
        value : Key
            The new value of the key.        
        """
        key = self.get_key_at(time)
        if key is not None:
            key.value = value
            self.keys.sort(key=lambda key: key.time)
        else:
            self.add_key(Key(time, value))

    def get_key_at(self, time):
        """Returns the key from the animation curve with the given time.

        Note
        ----
        If a key in the animation curve exists with a time within a tolerance of the given time,
        this key is returned.
        Tolerance values are defined in the constants.py file.
        Very high tolerance values are used here since it is never useful to have 2 keys extremely
        close to each other. They would become indistinguishable. See constants.py for more details.

        Parameters:
        ----------
        time : float     
            The time of the key to be returned.

        Returns:
        ----------
        Key
            The key at the given time. None if no key at the given time is found.
        """
        # Find the key close to the given time using binary search.
        low = 0
        high = len(self.keys) - 1
        while low <= high:
            mid = (low + high) // 2
            if self.keys[mid].time < (time - VERY_HIGH_A_TOL):
                low = mid + 1
            elif self.keys[mid].time > (time + VERY_HIGH_A_TOL):
                high = mid - 1
            else:
                return self.keys[mid]
        return None
    
    def remove_key_at(self, time):
        """Removes the key at the given time.

        Note:
        ----------
        this method uses the get_key_at() method to check if a key exists at the given time.
        See the documentation of that method for more details about the tolerance values of the check.
        
        Warns:
        ----------
        UserWarning
            If no key at the given time is found.
        
        Parameters:
        ----------
        time : float
            The time of the key to be removed.
        """
        key = self.get_key_at(time)
        if key is not None:
            self.keys.remove(key)
        else:
            warnings.warn("No key found at the given time. Key removal ignored.", stacklevel=2)
         
    def get_previous_key(self, time):
        """Returns the previous key of the animation curve at the given key time.
        i.e. the key with the largest time less than the given time.

        Parameters:
        ----------
        time : float
            The time at which the previous key is to be returned.

        Returns:
        ----------
        Key
            The previous key. None if no key is found.

        Note:
        ----------
        If the given time is less than the time of the first key in the animation curve,
        the first key is returned.
        """
        if not self.keys:
            return None
        
        # return the last key with a time less than the given time
        for key_id in range(len(self.keys) - 1, -1, -1):
            if self.keys[key_id].time <= time:
                return self.keys[key_id]

        return self.keys[0]       

    def get_next_key(self, time):
        """Returns the next key of the animation curve at the given key time.
        i.e. the key with the smallest time greater than the given time.

        Parameters:
        ----------
        time : float
            The time at which the next key is to be returned.
        
        Returns:
        ----------
        Key
          The next key. None if no key is found.

        Note:
        ----------
        If the given time is greater than the time of the last key in the animation curve,
        the last key is returned.
        """
        if not self.keys:
            return None

        for key_id in range(len(self.keys)):
            if self.keys[key_id].time > time:
                return self.keys[key_id]  

        return self.keys[-1]

    def get_previous_value(self, time):
        """Returns the previous value of the animation curve at the given key time.

        Parameters:
        ----------
        time : float
            The time at which the previous value is to be returned.

        Returns:
        ----------
        Any
            The previous value. None if no key is found.

        See Also:
        ----------
        get_previous_key(): used to get the previous key from which the value is returned.

        Note
        ----
        Used for constant interpolation.
        """
        # Check if keys list is empty
        if self.get_previous_key(time) is None:
            return None

        previous_key = self.get_previous_key(time)
        return previous_key.value

    def get_lerp_value(self, time):
        """Computes the lerp value for the given key time.

        The lerp value is the value obtained by linearly interpolating
        between the previous and next keys.
        Not all key types support lerp interpolation. See the 
        class description for a list of supported key types.

        Parameters:
        ----------
        time : float
            The time where to compute the lerp value.

        Returns:
        ----------
        any
            The lerp value. None if no key is found.

        See Also:
        ----------
        get_previous_key(): used to get the previous key.
        get_next_key(): used to get the next key.
        lerp(): used to compute the lerp value.

        Note
        ----
        Used for linear interpolation.
        """
        # Check if keys list is empty
        if self.get_previous_key(time) is None:
            return None

        previous_key = self.get_previous_key(time)
        next_key = self.get_next_key(time)
        delta = next_key.time - previous_key.time

        if np.isclose(delta, 0, rtol=VERY_HIGH_R_TOL, atol=VERY_HIGH_A_TOL):
            return previous_key.value
        else:
            progress = (time - previous_key.time) / delta
            pk_val = previous_key.value
            nk_val = next_key.value

            return Curve.lerp(pk_val, nk_val, progress)

    @staticmethod
    def lerp(first_val, second_val, factor):
        """Computes the lerp value at the given factor between two values.
        
        A factor of 0 returns the first value.
        Similarly, a factor of 1 returns the second value.

        For int, float and array of int and float, The interp() function
        of numpy is used to compute the lerp value.
        See https://numpy.org/doc/stable/reference/generated/numpy.interp.html
        For quaternion, the slerp() function of numpy quaternion is used.
        See https://github.com/moble/quaternion/blob/main/src/quaternion/quaternion_time_series.py#L61
        For transform, the lerp() function of the Transform class is used.
        This method makes use of the two previous methods to compute the lerp value.
        See skeletal_animation/core/math/transform.py
        All other types are not supported.

        Parameters:
        ----------
        first_val : any
            The first value.
        second_val : any
            The second value.
        factor : float
            The factor used to compute the lerp value.
        
        Returns:
        ----------
        any
            The linearly interpolated value.

        Raises:
        ----------
        TypeError
            If the type of the given values is not supported.
            See the class description for a list of supported types.

        Note:
        ----------
        The type checking is only done on the first_val parameter.
        The type of the two values must be the same.
        """
        if isinstance(first_val, (int, float)):
            return np.interp(factor, [0, 1], [first_val, second_val])

        elif isinstance(first_val, (list, np.ndarray)):
            size = len(first_val)
            v = np.empty(size)
            for i in range(len(first_val)):
                v[i] = np.interp(factor, [0, 1], [first_val[i], second_val[i]])
            return v

        elif isinstance(first_val, np.quaternion):
            return quaternion.slerp(first_val, second_val, 0, 1, factor)

        elif isinstance(first_val, Transform):
            return Transform.lerp(first_val, second_val, factor)
        else:
            raise TypeError(f"Cannot lerp between elements of type {str(type(first_val))}.")

    def __repr__(self):
        if self.keys:
            return f"Curve(type={str(self.keys[0].get_type())}, keys={str(self.keys)})"
        else:
            return "Curve(type=None, keys=[])"

    def __eq__(self, other):
        """Equality operator for Curve objects.

        Note
        ----
        Two curves are considered equal if all keys are equal. See Key.__eq__()

        Parameters
        ----------
        other : Curve
            Curve object to compare to.

        Returns
        -------
        bool
            True if the two Curves are equal, False otherwise.
        """
        return (np.all(self.keys == other.keys))


from setuptools import setup, find_packages

setup(name='skeletal_animation',
      version='0.1.1',
      description='Standalone package for skeletal animation',
      long_description="# This is a work in progrees",
      long_description_content_type='text/markdown',
      url='https://gitlab-etu.ing.he-arc.ch/isc/2021-22/niveau-3/tb-iscdlm/226/tb-animation-squelettale',
      author='Dimitri "Kerzoum" Kohler',
      author_email='dimitri.kohler@he-arc.ch',      
      license='MIT',
      classifiers=[
            # How mature is this project? Common values are
            #   3 - Alpha
            #   4 - Beta
            #   5 - Production/Stable
            'Development Status :: 3 - Alpha',

            # Indicate who your project is intended for
            'Intended Audience :: Developers',
            'Topic :: Multimedia :: Graphics',

            # Pick your license as you wish (should match "license" above)
            'License :: OSI Approved :: MIT License',

            # Specify the Python versions you support here. In particular, ensure
            # that you indicate whether you support Python 2, Python 3 or both.
            'Programming Language :: Python :: 3.7'
            ],
      packages=find_packages(),
      package_data={
            'skeletal_animation': ['animated_models/*.pkl', 'examples/*', 'gui/*', 'gui/assets/shaders/*.glsl'],
      },
      include_package_data=True,
)


import warnings
import numpy as np
import pickle

from skeletal_animation.core.math import *
from skeletal_animation.core.model import *
from skeletal_animation.core.animate import *


def serialize(animator, file_name, path):
    """Serialize the current animator to a file using pickle.

    Animator object is saved as a pickle file (.pkl).

    Parameters:
    -----------
    animator : Animator
        The animator to serialize
    file_name : str
        The name of the file to save. Extension .pkl is added automatically.
    path : str
        The path to where the file will be saved.
    """
    with open(path + file_name + ".pkl", 'wb') as f:
        pickle.dump(animator, f)


def load_serialized(file_name, path):
    """Load a serialized animator from a file using pickle.

    Animator object is loaded from a pickle file (.pkl).

    Parameters:
    -----------
    file_name : str
        The name of the file to load. Extension .pkl is added automatically.
    path : str
        The path to the file to load.

    Returns:
    --------
    animator : Animator
        The loaded Animator object.

    Raises:
    -------
    FileNotFoundError
        If the file cannot be found.
    """
    try:
        with open(path + file_name + ".pkl", 'rb') as f:
            return pickle.load(f)
    except FileNotFoundError:
        raise FileNotFoundError("File not found: " + path + file_name + ".pkl")


def load_fbx(fbx_file_path):
    """Creates an Animator object from a fbx file.

    This method requires the installation of the fbx sdk for python.
    The fbx sdk is available at: https://www.autodesk.com/developer-network/platform-technologies/fbx-sdk-2020-2-1.

    Parameters:
    -----------
    fbx_file_path : str
        The path to the fbx file to load.

    Raises:
    -------
    ImportError
        If the FBX SDK is not installed.

    Returns:
    --------
    animator : Animator
        The generated Animator object.

    Warnings
    --------
    This loader does not support all fbx files.
    Fbx files must follow the following criteria:
    - Animation data contains only one animation stack with only one animation layer.
    - If a key contains a value for one axis, the values for all other axes must be specified.
    Fbx files exported from Mixamo follow these criteria.
    """
    # check if the fbx sdk is installed 
    try:
        import fbx
    except ImportError:
        warnings.warn("The FBX SDK is not installed.\
            To load animation files from fbx, the FBX SDK must be installed.\
            see https://www.autodesk.com/developer-network/platform-technologies/fbx-sdk-2020-0",
            stacklevel=2)

    sdk_manager = fbx.FbxManager.Create()

    ios = fbx.FbxIOSettings.Create(sdk_manager, fbx.IOSROOT)
    sdk_manager.SetIOSettings(ios)

    importer = fbx.FbxImporter.Create(sdk_manager, '')

    if not importer.Initialize(fbx_file_path, -1, sdk_manager.GetIOSettings()):
        print("Call to FbxImporter::Initialize() failed.")
        raise ImportError("Error importing file %s file not found or %s" % (
            fbx_file_path, importer.GetStatus().GetErrorString(),
            ))

    scene = fbx.FbxScene.Create(sdk_manager, "myScene")
    importer.Import(scene)
    importer.Destroy()

    # _____________LOADING FBX HIERARCHY_____________
    def load_skeleton_nodes(skeleton_node_list, root_node):
        for i in range(root_node.GetChildCount()):
            child_node = root_node.GetChild(i)
            for i in range(child_node.GetNodeAttributeCount()):
                attribute = child_node.GetNodeAttributeByIndex(i)
                if type(attribute) == fbx.FbxSkeleton:
                    skeleton_node_list.append(child_node)
            load_skeleton_nodes(skeleton_node_list, child_node)

    def get_anim_layer():
        nb_anim_stack = scene.GetSrcObjectCount(
            fbx.FbxCriteria.ObjectType(fbx.FbxAnimStack.ClassId)
            )

        if nb_anim_stack == 0:
            raise ValueError("No animation stack found in fbx file.")
        elif nb_anim_stack > 1:
            warnings.warn("Multiple anim stacks found in fbx file, only the first one is treated", stacklevel=2)

        anim_stack = scene.GetSrcObject(
            fbx.FbxCriteria.ObjectType(fbx.FbxAnimStack.ClassId)
            , 0)

        nb_anim_layers = anim_stack.GetSrcObjectCount(
            fbx.FbxCriteria.ObjectType(fbx.FbxAnimLayer.ClassId)
            )

        if nb_anim_layers == 0:
            raise ValueError("No animation layer found in anim stack.")
        elif nb_anim_layers > 1:
            warnings.warn("Multiple anim layers found in anim stack, only the first one is treated", stacklevel=2)

        anim_layer = anim_stack.GetSrcObject(
            fbx.FbxCriteria.ObjectType(fbx.FbxAnimLayer.ClassId)
            , 0)

        return anim_layer

    def load_curves(skeleton_node, anim_layer):
        # Assumes that when a change to the position/orientation of a joint is made,
        # the translation/rotation values for each axis are given in the keyframe.

        tX_curve = skeleton_node.LclTranslation.GetCurve(anim_layer, "X")
        tY_curve = skeleton_node.LclTranslation.GetCurve(anim_layer, "Y")
        tZ_curve = skeleton_node.LclTranslation.GetCurve(anim_layer, "Z")
        rX_curve = skeleton_node.LclRotation.GetCurve(anim_layer, "X")
        rY_curve = skeleton_node.LclRotation.GetCurve(anim_layer, "Y")
        rZ_curve = skeleton_node.LclRotation.GetCurve(anim_layer, "Z")

        translation_keys = []
        rotation_keys = []

        duration = 1.0

        if tX_curve is not None and tY_curve is not None and tZ_curve is not None:
            for key_id in range(tX_curve.KeyGetCount()):
                key_time = tX_curve.KeyGetTime(key_id).GetTimeString("")

                t = [tX_curve.KeyGetValue(key_id), tY_curve.KeyGetValue(key_id), tZ_curve.KeyGetValue(key_id)]

                if "*" in key_time:
                    # Keys whose time is marked with "*" are not used in the animation.
                    # They are probably used to improve interpolation quality in between keyframes.         
                    pass
                else:
                    key_time = float(key_time)
                    if key_time > duration:
                        duration = key_time

                    translation_keys.append(Key(key_time, t))

        if rX_curve is not None and rY_curve is not None and rZ_curve is not None:
            for key_id in range(rX_curve.KeyGetCount()):
                key_time = rX_curve.KeyGetTime(key_id).GetTimeString("")

                v = fbx.FbxVector4(rX_curve.KeyGetValue(key_id), rY_curve.KeyGetValue(key_id), rZ_curve.KeyGetValue(key_id))

                m = fbx.FbxAMatrix()
                m.SetR(v)
                q = m.GetQ()
                q = np.quaternion(q.GetAt(3), q.GetAt(0), q.GetAt(1), q.GetAt(2))

                if "*" in key_time:
                    # Keys whose time is marked with "*" are not used in the animation.
                    # They are probably used to improve interpolation quality in between keyframes.
                    pass
                else:
                    key_time = float(key_time)
                    if key_time > duration:
                        duration = key_time

                    rotation_keys.append(Key(key_time, q))

        for key in translation_keys:
            key.time /= duration
        for key in rotation_keys:
            key.time /= duration

        translation_curve = Curve(translation_keys)
        rotation_curve = Curve(rotation_keys)
        return translation_curve, rotation_curve, duration

    # Assumes that the animation data contains only one animation stack with only one animation layer.
    # Mixamo rigged animations have this structure.
    # Assumes the duration of the animation for each joint is the same.

    root_node = scene.GetRootNode()
    skeleton_node_list = []

    load_skeleton_nodes(skeleton_node_list, root_node)

    skeleton_root = None
    root_set = False

    anim_layer = get_anim_layer()
    joint_animations = []

    joint_hierarchy = []

    anim_duration = 1.0

    for skeleton_node in skeleton_node_list:
        # Creating joints
        joint_name = skeleton_node.GetName()

        pos = [float(val) for val in skeleton_node.LclTranslation.Get()]

        rotation = skeleton_node.PreRotation.Get()

        v = fbx.FbxVector4(rotation)

        m = fbx.FbxAMatrix()
        m.SetR(v)
        q = m.GetQ()        
        orient = np.quaternion(q.GetAt(3), q.GetAt(0), q.GetAt(1), q.GetAt(2))

        joint = Joint(name=joint_name, local_bind_transform=Transform(pos, orient))

        children = []
        for i in range(skeleton_node.GetChildCount()):
            child_node = skeleton_node.GetChild(i)
            children.append(child_node.GetName())

        joint_hierarchy.append((joint, children))

        if not root_set:
            skeleton_root = joint
            root_set = True

        # Creating joint animations
        if anim_layer:
            translation_curve, rotation_curve, duration = load_curves(skeleton_node, anim_layer)
            if duration > anim_duration:
                anim_duration = duration
            joint_animation = JointAnimation(joint, translation_curve, rotation_curve)
            joint_animations.append(joint_animation)

    # Building joint hierarchy
    joint_list = [relation[0] for relation in joint_hierarchy]

    for relation in joint_hierarchy:
        children = [joint for joint in joint_list if joint.name in relation[1]]
        relation[0].add_children(children)

    # Creating skeleton
    skeleton = Skeleton(skeleton_root)

    # Creating animation
    animation = Animation(joint_animations, anim_duration)

    animator = Animator(animation=animation, skeleton=skeleton)

    return animator
